/*
 *	Network Block Device - server
 *
 *  copyright 1996, 97 Pavel Machek <pavel@atrey.karlin.mff.cuni.cz>
 *  copyright 1998 Josh Parsons <josh@coombs.anu.edu.au>
 *
 */

/* 
 * I have fixed many things.  In particular:
 * - use stat() to find devsize, not lseek
 * - code cleanup
 * - read only exports are supported (good for cdroms)
 * - local IO can be done through mmap() (faster than lseek/read/write)
 * - endianness
 * TODO:
 * - support for being run by inetd
 * - restart (reopen export?) on SIGHUP
 * - shutdown politely (and inform client) on SIGTERM, SIGSEGV, SIGBUS
 * - allow user to specify length of device (and origin?)
 * - allow ramdisks with MAP_PRIVATE
 *         --- Josh
 */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <netinet/in.h>         /* sockaddr_in, htons, in_addr */
#include <netdb.h>              /* hostent, gethostby*, getservby* */
#include <syslog.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <linux/nbd.h>

int port;
int read_only=0;

int export=-1;
char *export_mmap=NULL;
size_t devsize=0;

#define DODBG
#define DEBUG( a ) printf( a )
//#define DEBUG( a ) 

void err( const char *s ) { 
   /* log the error */
   syslog( LOG_ERR, s ); 
   
   /* clean up */
   if(export_mmap) munmap(export_mmap,devsize);
   if(export>=0) close(export);

   /* die */
   exit( 1 ); 
}

void
readit( int f, void *buf, int len ) {
   int res;
   while( len>0 )
      {
	 DEBUG( "*" );
	 res = read( f, buf, len );
	 if (res <= 0) err( "Read failed: %m" );
	 len -= res;
	 buf += res;
      }
}

void
writeit( int f, void *buf, int len ) {
   int res;
   while( len>0 )
      {
	 DEBUG( "+" );
	 res = write( f, buf, len );
	 if (res <= 0) err( "Write failed: %m" );
	 len -= res;
	 buf += res;
      }
}


int 
main(int argc, char *argv[]) {
   struct sockaddr_in addrin;
   int addrinlen = sizeof( addrin );
   int sock, net;
   struct nbd_request request;
   struct nbd_reply reply;
   unsigned int magic = 0x81861253;
   unsigned int j;
   int size = 256;
   int i=0;

#ifdef DODBG
   setvbuf( stdout, NULL, _IONBF, 0 );
   setvbuf( stderr, NULL, _IONBF, 0 );
#endif

   if (argc < 3 || argc > 5)  { 
      printf( "Usage: port file_to_export [size] [-ro]\n" ); 
      exit( 0 ); 
   }

   port = atoi( argv[1] ); 
   if(argc>3&&!strcmp(argv[3],"-ro")) read_only=1;
   if(argc>4&&!strcmp(argv[4],"-ro")) read_only=1;
   if(argc>3) devsize=atoi(argv[3]);
   export = open( argv[2], read_only?O_RDONLY:O_RDWR );

   openlog("nbd_server", LOG_PID, LOG_DAEMON);
   
   if (export == -1) err( "Could not open exported file: %m" );
   if(devsize<=0) {
#if 0
      devsize = lseek(export, 0, SEEK_END);
      if (devsize == -1) err( "Could not find size of exported file: %m" );
#else
      struct stat st;
      if(fstat(export,&st)) err( "Could not fstat exported file: %m" );
      devsize=st.st_size;
#endif
   };
   printf( "Device size recognized as %d.\n", devsize );

   /* now try to mmap() the export */
   export_mmap=mmap(NULL,devsize,
		    read_only?PROT_READ:(PROT_READ|PROT_WRITE),
		    MAP_SHARED,
		    export,0);
   if(export_mmap==(caddr_t)-1||export_mmap==NULL) {
      DEBUG("mmap() failed.  Will use lseek/read/write instead.\n");
      export_mmap=NULL;
   };
   
   if ( (sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP))<0 ) 
      err( "socket: %m"); 
   DEBUG( "Setsockopt..." );
   setsockopt( sock, SOL_SOCKET, SO_SNDBUF, &size, sizeof( int ));
   size =1;
   setsockopt( sock, SOL_TCP, TCP_NODELAY, &size, sizeof( int ));
   size =1024;
   setsockopt( sock, SOL_TCP, TCP_MAXSEG, &size, sizeof( int ));
   DEBUG( " ok!\n" );
   
   DEBUG( "Waiting for connections... bind, " );
   addrin.sin_family = AF_INET;
   addrin.sin_port = htons( port );
   addrin.sin_addr.s_addr = 0;
   if (bind( sock, (struct sockaddr *) &addrin, addrinlen )<0) 
      err( "bind: %m"); 
   DEBUG( "listen, " );
   if (listen( sock, 1 )<0) err( "listen: %m"); 
   DEBUG( "accept, " );
   if ((net = accept( sock, (struct sockaddr *) &addrin, &addrinlen ))<0) 
      err( "accept: %m" ); 
   
   /* Trouble here.  Pavel's version used to send magic and devsize
      in host order.  I have changed this to network order.  So that 
      horrors do not ensue when my client tries to talk to his server,
      or vice versa, I have reversed the byte order of the magic number.
      This means that the versions are inter-operable on machines where
      the orders are the same, but magic number detection will kick in
      where they are different */

   DEBUG( "Negotiating..." );
   if (write( net, "MAGICNBD", 8 )<0) 
      err( "Negotiation failed: %m\n" );
   j=htonl(magic);
   if (write( net, &j, sizeof( j ) )<0) 
      err( "Negotiation failed: %m\n" );
   j=htonl(devsize);
   if (write( net, &j, sizeof( j ) )<0) 
      err( "Negotiation failed: %m\n" ); 
   DEBUG( "done!\n" );
   
   DEBUG( "Entering request loop!\n" );
   while (1)
      {
	 char buf[ 10240 ];
	 int len;
	 i++;
	 readit( net, &request, sizeof( request ));
	 request.len = ntohl( request.len );
	 request.from = ntohl( request.from );
	 request.magic = ntohl( request.magic );
	 request.type = ntohl( request.type );
#ifdef DODBG
	 printf( "%d: ", i );
#endif
	 if (request.magic != NBD_REQUEST_MAGIC) err( "Not enough magic.\n" );
	 len = request.len;
	 if (len>10240) err( "Request too big!\n" );
#ifdef DODBG
	 printf( "%s from %d (%d) len %d, ", 
		 (request.type ? "WRITE" : "READ"), 
		 request.from, request.from / 512, len );
#endif
	 reply.magic = NBD_REPLY_MAGIC;
	 memcpy( reply.handle, request.handle, sizeof( reply.handle ));
	 reply.error = 0;

	 /* check for problems with request */
	 if ((request.from<0) || ((request.from + request.len)>devsize)) { 
	    DEBUG( "[RANGE!]" ); 
	    reply.error=1; 
	 }
	 if ((request.type) && read_only) {
	    DEBUG( "[WRITE PROTECT!]");
	    reply.error=1;
	 }

	 /* deal with a write request */
	 if ((request.type) && (!reply.error))	/* WRITE */
	    {
	       /* if we have a mmap, use that */
	       if(export_mmap) {
		  readit(net, export_mmap+request.from, len);
		  DEBUG("wr: net->exp, "); 
	       }
	       /* else do it the old way */
	       else {
		  DEBUG( "wr: net->buf, " );
		  readit( net, buf, len );
		  if (lseek( export, request.from, SEEK_SET )<0) 
		     err( "Can not seek locally!\n" );
		  DEBUG( "buf->exp, " );
		  writeit( export, buf, len );
	       }
	    }

	 /* set up to reply */
	 DEBUG( "replying, " );
	 reply.magic = htonl( reply.magic );
	 reply.error = htonl( reply.error );
	 writeit( net, &reply, sizeof( reply ));

	 /* deal with a read request */
	 if (!request.type) /* READ */
	    {
	       if (!reply.error)
		  {
		     /* use the mmap if we have it */
		     if(export_mmap) {
			DEBUG( "exp->net, " );
			writeit( net, export_mmap+request.from, len );
		     }
		     /* do it the old way */
		     else {
			if (lseek( export, request.from, SEEK_SET )<0) 
			   err( "Can not seek locally /2!\n" );
			DEBUG( "exp->buf, " );
			if (read( export, buf, len )!=len) { 
			   DEBUG( "Read failed: %m" ); 
			   reply.error = 1; 
			}
			DEBUG( "buf->net, " );
			writeit( net, buf, len );
		     }
		  }
	    }

	 /* all done */
	 DEBUG( "OK!\n" );
      }
}


